import argparse
import asyncio
import sys

import pyaudio

from aiovban import VBANApplicationData, DeviceType
from aiovban.asyncio import AsyncVBANClient
from aiovban.enums import Features
from ..util import get_device_by_name, setproctitle
from ... import VBANAudioPlayer, __version__


def setup_logging():
    import logging

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    print("Setting up logging at level INFO")
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    root.addHandler(handler)


async def run_loop(config):
    application_data = VBANApplicationData(
        application_name="VBAN Audio Sender",
        features=Features.Audio | Features.Text,
        device_type=DeviceType.Receptor,
        version=__version__,
    )
    client = AsyncVBANClient(application_data=application_data)

    full_address, stream_name = config.stream.split("/")
    if ":" in full_address:
        address, port = full_address.split(":")
        port = int(port)
    else:
        address = full_address
        port = 6980

    device = await client.register_device(address, port)
    stream = await device.text_stream(stream_name)

    await stream.send_text(config.command)


def main():
    parser = argparse.ArgumentParser(
        prog="aioVBAN Text Sender",
        description="Receives Audio Streams from VBAN and plays them back",
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "--debug",
        action="store_false",
        help="Enable debug logging",
    )
    parser.add_argument(
        "-s",
        "--stream",
        help="Stream in the format of 'address:port/stream_name'",
    )
    parser.add_argument(
        "command",
        help="Command to run",
    )

    config = parser.parse_args()

    setproctitle("aioVBAN Text Sender")
    setup_logging()
    asyncio.run(run_loop(config))



import asyncio
import functools
import threading


def run_on_background_thread(func):
    def run_loop(loop, future, origin_loop, *args, **kwargs):
        try:
            asyncio.set_event_loop(loop)
            kwargs["origin_loop"] = origin_loop
            loop.run_until_complete(func(*args, **kwargs))
        except Exception as e:
            if not future.done():
                origin_loop.call_soon_threadsafe(future.set_exception, e)
        finally:
            loop.stop()
            loop.close()
            if not future.done():
                origin_loop.call_soon_threadsafe(future.set_result, None)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        loop = asyncio.new_event_loop()
        origin_loop = asyncio.get_running_loop()
        future = asyncio.get_running_loop().create_future()
        thread = threading.Thread(
            target=run_loop,
            args=(loop, future, origin_loop, *args),
            kwargs=kwargs,
            daemon=True,
        )
        thread.start()
        future.add_done_callback(lambda s: thread.join(timeout=1))
        return future

    return wrapper


class FrameBuffer:
    """
    A utility class to manage a buffer of audio frames. This class is used to store audio frames
    with a maximum frame count. When the buffer is full, the oldest frames are dropped to make
    room for new frames and keep latency low.

    Methods:
        __init__(max_frame_count: int, bytes_per_frame: int = 4):
            Initializes the FrameBuffer with a maximum frame count and bytes per frame.

        write(data: bytes, frames: int):
            Writes data to the buffer and updates the frame count. Thread-safe.

        size():
            Returns the current size of the buffer in bytes and the frame count. Thread-safe.

        read(num_frames: int):
            Reads a specified number of frames from the buffer, dropping the oldest frames if necessary. Thread-safe.

        synchronize(bytes_per_frame: int):
            Resets the buffer and updates the bytes per frame. Thread-safe.
    """

    def __init__(self, max_frame_count: int, bytes_per_frame: int = 1):
        """
        Initializes the FrameBuffer with a maximum frame count and bytes per frame.

        Args:
            max_frame_count (int): The maximum number of frames the buffer can hold.
            bytes_per_frame (int): The number of bytes per frame. Default is 4.
        """
        self._buffer = bytearray()
        self._frame_count = 0
        self._max_frame_count = max_frame_count
        self._bytes_per_frame = bytes_per_frame
        self._mutex = threading.Lock()

    def write(self, data: bytes, frames: int):
        """
        Writes data to the buffer and updates the frame count.

        Args:
            data (bytes): The audio data to write to the buffer.
            frames (int): The number of frames in the data.
        """
        with self._mutex:
            self._buffer.extend(data)
            self._frame_count += frames

    def size(self):
        """
        Returns the current size of the buffer in bytes and the frame count.

        Returns:
            tuple: A tuple containing the size of the buffer in bytes and the frame count.
        """
        with self._mutex:
            return len(self._buffer), self._frame_count

    def read(self, num_frames: int, drop_frames=True):
        """
        Reads a specified number of frames from the buffer, dropping the oldest frames if necessary.

        Args:
            num_frames (int): The number of frames to read from the buffer.
            drop_frames (bool): Whether to drop frames if the buffer is full. Default is True.

        Returns:
            tuple: A tuple containing the read data, the number of frames read, and the number of frames dropped.
        """
        with self._mutex:
            if drop_frames:
                excess_frames = max(0, self._frame_count - self._max_frame_count)
                if excess_frames > 0:
                    bytes_to_drop = self._bytes_per_frame * excess_frames
                    del self._buffer[:bytes_to_drop]
                    self._frame_count -= excess_frames
            else:
                excess_frames = 0

            maximum_available_frames = min(num_frames, self._frame_count)
            bytes_for_frames = self._bytes_per_frame * maximum_available_frames

            buffer_data = bytes(self._buffer[:bytes_for_frames])
            del self._buffer[:bytes_for_frames]
            self._frame_count -= maximum_available_frames

            return buffer_data, maximum_available_frames, excess_frames

    def synchronize(self, bytes_per_frame: int):
        """
        Resets the buffer and updates the bytes per frame.

        Args:
            bytes_per_frame (int): The new number of bytes per frame.
        """
        with self._mutex:
            self._buffer.clear()
            self._frame_count = 0
            self._bytes_per_frame = bytes_per_frame



import argparse
import asyncio
import logging
import sys

import pyaudio

from aiovban import VBANApplicationData, DeviceType, VBANSampleRate
from aiovban.asyncio import AsyncVBANClient
from aiovban.enums import Features
from aiovban.asyncio.util import BackPressureStrategy
from ..util import get_device_by_name, setproctitle
from ... import VBANAudioPlayer, __version__

logger = logging.getLogger(__name__)


def setup_logging(debug=False):
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG if debug else logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    root.addHandler(handler)


async def wait_for_first_done(*tasks):
    # Wrap coroutines in tasks, as asyncio.wait() requires tasks/futures since Python 3.11
    wrapped = [asyncio.ensure_future(t) for t in tasks]
    (done_tasks, pending) = await asyncio.wait(
        wrapped, return_when=asyncio.FIRST_COMPLETED
    )
    for pending_task in pending:
        pending_task.cancel()
    return done_tasks


async def run_loop(config):
    application_data = VBANApplicationData(
        application_name="VBAN Audio Receiver",
        features=Features.Audio | Features.Text,
        device_type=DeviceType.Receptor,
        version=__version__,
    )
    client = AsyncVBANClient(application_data=application_data)
    listen_future = await client.listen(config.host_address, config.host_port)

    pyaudio_instance = pyaudio.PyAudio()

    output_device = get_device_by_name(pyaudio_instance, config.output_device)

    players = []
    for stream in config.streams:
        full_address, stream_name = stream.split("/")
        if ":" in full_address:
            address, port = full_address.split(":")
            port = int(port)
        else:
            address = full_address
            port = 6980

        host = await client.register_device(address, port)
        receiver = host.receive_stream(
            stream_name, back_pressure_strategy=BackPressureStrategy.DRAIN_OLDEST
        )
        players.append(
            VBANAudioPlayer(
                stream=receiver,
                pyaudio=pyaudio_instance,
                device_index=output_device,
                channels=config.channels,
                sample_rate=VBANSampleRate.find(config.sample_rate),
            )
        )

    await wait_for_first_done(listen_future, *map(lambda p: p.listen(), players))


def main():
    parser = argparse.ArgumentParser(
        prog="aioVBAN Stream Receiver",
        description="Receives Audio Streams from VBAN and plays them back",
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug logging",
    )
    parser.add_argument(
        "--host-port",
        type=int,
        default=6980,
    )
    parser.add_argument("--host-address", default="0.0.0.0")
    parser.add_argument(
        "streams",
        nargs="+",
        help="Streams in the format of 'address:port/stream_name'",
    )
    parser.add_argument(
        "--output-device",
        type=str,
        help="The name of the output device to use",
    )
    parser.add_argument(
        "--channels",
        type=int,
        default=2,
        help="Initial number of channels to use",
    )
    parser.add_argument(
        "--sample-rate",
        type=int,
        default=48000,
        help="Initial sample rate to use",
    )

    config = parser.parse_args()
    setup_logging(config.debug)

    try:
        import uvloop
        uvloop.install()
        logger.info("Using uvloop for high-performance asyncio")
    except ImportError:
        logger.info("uvloop not found, using standard asyncio event loop")

    setproctitle("aioVBAN Stream Receiver")
    asyncio.run(run_loop(config))



import argparse
import asyncio
import logging
import sys

import pyaudio

from aiovban import VBANApplicationData, DeviceType, VBANSampleRate
from aiovban.asyncio import AsyncVBANClient
from aiovban.enums import Features
from ..util import get_device_by_name, setproctitle
from ... import VBANAudioSender, __version__

logger = logging.getLogger(__name__)


def setup_logging(debug=False):
    import logging

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG if debug else logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    root.addHandler(handler)


async def run_loop(config):
    application_data = VBANApplicationData(
        application_name="VBAN Audio Sender",
        features=Features.Audio | Features.Text,
        device_type=DeviceType.Transmitter,
        version=__version__,
    )
    client = AsyncVBANClient(application_data=application_data)

    pyaudio_instance = pyaudio.PyAudio()

    input_device = get_device_by_name(pyaudio_instance, config.input_device)

    device = await client.register_device(config.address, config.port)
    logger.info(f"Registered device {device}")
    stream = await device.send_stream(config.stream_name)

    sample_rate = VBANSampleRate.find(config.sample_rate)
    if not sample_rate:
        logger.error(f"Invalid sample rate: {config.sample_rate}")
        return

    listener = VBANAudioSender(
        stream=stream,
        pyaudio=pyaudio_instance,
        device_index=input_device,
        channels=config.channels,
        sample_rate=VBANSampleRate.find(config.sample_rate),
        framebuffer_size=config.framebuffer_size,
    )
    await listener.listen()


def main():
    parser = argparse.ArgumentParser(
        prog="aioVBAN Stream Sender",
        description="Captures audio from an input device and sends it via VBAN",
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug logging",
    )
    parser.add_argument(
        "--address",
        type=str,
        help="The address to send the VBAN packets to",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=6980,
    )
    parser.add_argument(
        "--input-device",
        type=str,
        help="The name of the input device to use",
    )
    parser.add_argument(
        "--stream-name",
        type=str,
        help="The name of the stream to send",
    )
    parser.add_argument(
        "--channels",
        type=int,
        default=2,
        help="Number of channels to use",
    )
    parser.add_argument(
        "--sample-rate",
        type=int,
        default=48000,
        help="Sample rate to use",
    )
    parser.add_argument(
        "--framebuffer-size",
        type=int,
        default=256,
        help="Number of frames per VBAN packet",
    )

    config = parser.parse_args()
    setup_logging(config.debug)

    try:
        import uvloop
        uvloop.install()
        logger.info("Using uvloop for high-performance asyncio")
    except ImportError:
        logger.info("uvloop not found, using standard asyncio event loop")

    setproctitle("aioVBAN Stream Sender")
    asyncio.run(run_loop(config))



from enum import Enum

import pyaudio

from aiovban.packet.headers.audio import BitResolution


class VBANPyAudioFormatMapping(Enum):
    BYTE8 = BitResolution.BYTE8, pyaudio.paUInt8
    INT16 = BitResolution.INT16, pyaudio.paInt16
    INT24 = BitResolution.INT24, pyaudio.paInt24
    INT32 = BitResolution.INT32, pyaudio.paInt32
    FLOAT32 = BitResolution.FLOAT32, pyaudio.paFloat32

    def __new__(cls, data, pyaudio_format):
        obj = object.__new__(cls)
        obj.key = obj._value_ = data
        obj.pyaudio_format = pyaudio_format
        return obj



import importlib.metadata

from .enums import VBANPyAudioFormatMapping
from .player import VBANAudioPlayer
from .sender import VBANAudioSender

try:
    __version__ = importlib.metadata.version("aiovban-pyaudio")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"



import asyncio
import logging
from asyncio import AbstractEventLoop
from dataclasses import field, dataclass
from typing import Any

import pyaudio

from aiovban import VBANSampleRate
from aiovban.asyncio.streams import VBANOutgoingStream, BufferedVBANOutgoingStream
from aiovban.packet import VBANPacket, BytesBody
from aiovban.packet.headers.audio import VBANAudioHeader, BitResolution, Codec
from .enums import VBANPyAudioFormatMapping
from .util import run_on_background_thread

logger = logging.getLogger(__package__)


@dataclass
class VBANAudioSender:
    stream: VBANOutgoingStream
    device_index: int = 0

    sample_rate: VBANSampleRate = VBANSampleRate.RATE_48000
    channels: int = 2
    format: BitResolution = BitResolution.INT16
    framebuffer_size: int = 128
    sample_buffer_size: int = 3

    pyaudio: Any = field(default_factory=pyaudio.PyAudio, repr=False)
    _stream: Any = field(init=False, repr=False)
    _loop: Any = field(default=None, init=False, repr=False)
    _running: bool = field(default=False, init=False, repr=False)
    _sent_packet_count: int = field(default=0, init=False, repr=False)

    def __post_init__(self):
        self._stream = self.setup_stream()

    @property
    def bytes_per_frame(self):
        return self.channels * self.format.byte_width

    def setup_stream(self):
        return self.pyaudio.open(
            format=VBANPyAudioFormatMapping(self.format).pyaudio_format,
            channels=self.channels,
            rate=self.sample_rate.rate,
            input=True,
            input_device_index=self.device_index,
            frames_per_buffer=self.framebuffer_size,
        )

    def split_bytes_into_chunks(self, data, chunk_size):
        """Splits bytes into chunks of a given size."""
        return [data[i : i + chunk_size] for i in range(0, len(data), chunk_size)]

    def pack_audio_data(self, audio_data):
        samples_per_frame = len(audio_data) // self.bytes_per_frame
        packet = VBANPacket(
            header=VBANAudioHeader(
                streamname=self.stream.name,
                sample_rate=self.sample_rate,
                codec=Codec.PCM,
                channels=self.channels,
                bit_resolution=self.format,
                samples_per_frame=samples_per_frame,
            ),
            body=BytesBody(audio_data),
        )
        if isinstance(self.stream, BufferedVBANOutgoingStream):
            # Safe cross-thread async scheduling
            self.stream.send_packet_threadsafe(packet, loop=self._loop)
            self._sent_packet_count += 1
        elif self._loop:
            # Fallback for non-buffered streams
            asyncio.run_coroutine_threadsafe(self.stream.send_packet(packet), self._loop)
            self._sent_packet_count += 1

        if self._sent_packet_count % 100 == 0:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Sent {self._sent_packet_count} packets")

    def send_all_audio_data(self, audio_data):
        chunk_size = self.framebuffer_size * self.bytes_per_frame
        chunks = self.split_bytes_into_chunks(audio_data, chunk_size)
        for chunk in chunks:
            self.pack_audio_data(chunk)

    def read_stream(self, amount):
        return self._stream.read(amount, exception_on_overflow=False)

    @run_on_background_thread
    async def listen(self, origin_loop: AbstractEventLoop):
        self._loop = origin_loop
        self._running = True
        self._stream.start_stream()

        try:
            while self._running:
                audio_data = self.read_stream(
                    self.framebuffer_size * self.sample_buffer_size
                )
                self.send_all_audio_data(audio_data)
        except Exception as e:
            logger.error(f"Error in audio sender: {e}")
        finally:
            self.stop()

    def stop(self):
        self._running = False
        if self._stream.is_active():
            self._stream.stop_stream()
        self._stream.close()



import asyncio
import logging
from dataclasses import field, dataclass
from typing import Any

import pyaudio

from aiovban import VBANSampleRate
from aiovban.asyncio.streams import VBANIncomingStream
from aiovban.packet import VBANPacket, VBANHeader
from aiovban.packet.headers.audio import VBANAudioHeader, BitResolution
from .enums import VBANPyAudioFormatMapping
from .scripts.util import ProbabilityFilter
from .util import FrameBuffer

logger = logging.getLogger(__name__)
probability_filter = ProbabilityFilter()
probability_filter.probability = 0.001
probability_logger = logging.getLogger(__name__)
probability_logger.addFilter(probability_filter)


@dataclass
class VBANAudioPlayer:
    stream: VBANIncomingStream  # This stream should be accessed from the original loop, not the background thread

    device_index: int = 0
    sample_rate: VBANSampleRate = VBANSampleRate.RATE_48000
    channels: int = 2
    format: BitResolution = BitResolution.INT16
    framebuffer_size: int = 512
    max_framebuffer_size: int = 8192

    pyaudio: Any = None
    _stream: Any = field(
        init=False
    )  # The audio stream should be accessed from the background thread
    _framebuffer: FrameBuffer = field(default=None, init=False, repr=False)

    def __post_init__(self):
        self._synced = False
        self._framebuffer = FrameBuffer(
            self.max_framebuffer_size, self.format.byte_width * self.channels
        )

        if not self.pyaudio:
            self.pyaudio = pyaudio.PyAudio()

    def setup_stream(self):
        return self.pyaudio.open(
            format=VBANPyAudioFormatMapping(self.format).pyaudio_format,
            channels=self.channels,
            rate=self.sample_rate.rate,
            output=True,
            frames_per_buffer=self.framebuffer_size,
            output_device_index=self.device_index,
            stream_callback=self.data_callback_in_thread,
        )

    def check_pyaudio(self, packet: VBANPacket):
        header = packet.header
        if not isinstance(header, VBANAudioHeader):
            return False

        if (
            header.sample_rate != self.sample_rate
            or header.channels != self.channels
            or header.bit_resolution != self.format
        ):
            print(
                f"Changing stream to {header.channels} channels, {header.sample_rate.rate} Hz, {header.samples_per_frame} samples per frame for stream {header.streamname}"
            )
            if logger.isEnabledFor(logging.INFO):
                logger.info("Stopping stream")
            self._stream.stop_stream()
            if logger.isEnabledFor(logging.INFO):
                logger.info("Closing stream")
            self._stream.close()

            self.channels = header.channels
            self.sample_rate = header.sample_rate
            self.format = header.bit_resolution

            if logger.isEnabledFor(logging.INFO):
                logger.info("Opening new stream")
            self._stream = self.setup_stream()
            if logger.isEnabledFor(logging.INFO):
                logger.info("Starting new stream")
            self._stream.start_stream()

            return True
        return False

    def silence(self, num_frames=0):
        if self.format == BitResolution.BYTE8:
            return b"\x80" * num_frames * self.channels
        return b"\x00" * num_frames * self.format.byte_width * self.channels

    def data_callback_in_thread(self, in_data, frame_count, time_info, status):
        (buffer_size, available_frame_count) = self._framebuffer.size()

        # Wait for a cushion of data before starting to avoid immediate underflow
        if not self._synced:
            if available_frame_count < frame_count * 2:  # Wait for 2 buffers worth
                return self.silence(num_frames=frame_count), pyaudio.paContinue
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    f"Cushion reached ({available_frame_count} frames), starting playback"
                )
            self._synced = True

        (buffer_data, available_frames) = self.commit_data(frame_count)
        if available_frames < frame_count:
            # Only log underflow occasionally to avoid flooding
            if probability_filter.filter(None):
                logger.warning(
                    f"Buffer underflow: {frame_count - available_frames} frames with latency of {self._estimated_latency(frame_count - available_frames):.2f} ms"
                )
            return (
                self.silence(num_frames=frame_count - available_frames) + buffer_data,
                pyaudio.paContinue,
            )
        return buffer_data, pyaudio.paContinue

    def _frames_to_byte_count(self, frames):
        return frames * self.channels * self.format.byte_width

    def _estimated_latency(self, frame_count):
        return (frame_count / self.sample_rate.rate) * 1000  # ms

    def commit_data(self, frame_count):
        (buffer_size, available) = self._framebuffer.size()
        if probability_logger.isEnabledFor(logging.INFO):
            probability_logger.info(
                f"Buffer size: {buffer_size} \n Current Latency: {self._estimated_latency(available)} ms"
            )

        (buffer_data, available_frames, dropped_frames) = self._framebuffer.read(
            frame_count
        )
        if dropped_frames > 0:
            if logger.isEnabledFor(logging.INFO):
                logger.info(
                    f"Dropping {dropped_frames} frames out of {frame_count} frames with maximum of {self.max_framebuffer_size} frames"
                )
        return buffer_data, available_frames

    def write_data(self, packet: VBANPacket):
        header: VBANAudioHeader = packet.header
        byte_count = (
            header.samples_per_frame * header.channels * header.bit_resolution.byte_width
        )
        # packet.body.pack() already returns a memoryview or bytes
        data = packet.body.pack()[:byte_count]
        self._framebuffer.write(data, header.samples_per_frame)

    def sync_buffers(self):
        self._framebuffer.synchronize(self.format.byte_width * self.channels)

    async def listen(self):
        self._stream: pyaudio.Stream = self.setup_stream()
        self._stream.start_stream()

        try:
            while True:
                # Wait for at least one packet
                packet = await self.stream.get_packet()
                resync = self.check_pyaudio(packet)
                if resync:
                    self.sync_buffers()
                self.write_data(packet)

                # Now drain any other packets that arrived while we were processing
                while (next_packet := self.stream.get_packet_nowait()) is not None:
                    resync = self.check_pyaudio(next_packet)
                    if resync:
                        self.sync_buffers()
                    self.write_data(next_packet)
        except asyncio.CancelledError as _:
            self.stop()

    def stop(self):
        self._stream.stop_stream()
        self._stream.close()



import logging
from random import SystemRandom

import pyaudio

try:
    from setproctitle import setproctitle as _setproctitle
except ImportError:
    _setproctitle = None


def setproctitle(title: str):
    if _setproctitle:
        _setproctitle(title)


def get_device_by_name(instance: pyaudio.PyAudio, name: str):
    for i in range(instance.get_device_count()):
        device_info = instance.get_device_info_by_index(i)
        if device_info["name"] == name:
            return i
    return None


class ProbabilityFilter(logging.Filter):
    probability = 0.10
    cryptogen = SystemRandom()

    def filter(self, record):
        return self.cryptogen.random() < self.probability
